import functools
from time import sleep
import logging


def retries(max_tries, should_retry, delay=1, backoff=2):
    """
    Decorator that implements exponential backoff retry logic. If you have
    a function that may fail, this decorator can catch the exception and retry at
    exponentially increasing intervals until the number of retries is exhausted.

    The should_retry parameter should be a function that takes and exception as an argument
    and returns True if the function should be retried or False to permanently fail.

    This is extremely useful when working with external APIs. There is a shortcut
    decorator for working with Google APIs, see :func:`google_api_retries`.
    """
    def dec(func):
        functools.wraps(func)

        def f2(*args, **kwargs):
            seconds = delay
            tries = range(max_tries)
            tries.reverse()

            for tries_remaining in tries:
                try:
                    return func(*args, **kwargs)

                except Exception as e:
                    logging.info("Caught %s with %s retries left" % (e, tries_remaining))
                    if tries_remaining > 0 and should_retry(e):
                        logging.info("Exception raised, retrying in %s seconds" % seconds)
                        sleep(seconds)
                        seconds *= backoff
                    else:
                        raise e
                else:
                    break
        return f2
    return dec



import logging
import inspect
from collections import namedtuple
from google.appengine.api import search as search_api
from google.appengine.ext import ndb
from .ndb import Behavior


def _datetime_coverter(n, v):
    date = search_api.DateField(name=n, value=v)
    iso = search_api.TextField(name=n + '_iso', value=v.isoformat())
    return date, iso


property_to_field_map = {
    ndb.IntegerProperty: lambda n, v: search_api.NumberField(name=n, value=v),
    ndb.FloatProperty: lambda n, v: search_api.NumberField(name=n, value=v),
    ndb.BooleanProperty: lambda n, v: search_api.AtomField(name=n, value='true' if v else 'false'),
    ndb.StringProperty: lambda n, v: search_api.TextField(name=n, value=v),
    ndb.TextProperty: lambda n, v: search_api.TextField(name=n, value=v),
    # BlobProperty explicitly unindexable
    ndb.DateTimeProperty: _datetime_coverter,
    ndb.DateProperty: lambda n, v: search_api.DateField(name=n, value=v),
    ndb.TimeProperty: lambda n, v: search_api.TextField(name=n, value=v.isoformat()),
    ndb.GeoPtProperty: lambda n, v: search_api.GeoField(name=n, value=search_api.GeoPoint(v.lat, v.lon)),
    # KeyProperty explicity unindexable
    # BlobKeyProperty explicitly unindexable
    ndb.UserProperty: lambda n, v: search_api.TextField(name=n, value=unicode(v)),
    # StructuredProperty explicitly unindexable
    # LocalStructuredProperty explicitly unindexable
    # JsonProperty explicitly unindexable
    # PickleProperty explicity unindexable
    # GenericProperty explicitly unindexable
    # ComputedProperty explicitly unindexable
}

non_repeatable_properties = (
    ndb.DateTimeProperty,
    ndb.DateProperty,
    ndb.TimeProperty,
    ndb.IntegerProperty,
    ndb.FloatProperty
)


class Searchable(Behavior):
    """
    Automatically indexes models during after_put into the App Engine Text Search API.
    """
    def after_put(self, instance):
        only = self.Model.Meta.search_fields if hasattr(self.Model.Meta, 'search_fields') else None
        exclude = self.Model.Meta.search_exclude if hasattr(self.Model.Meta, 'search_exclude') else None
        indexer = self.Model.Meta.search_indexer if hasattr(self.Model.Meta, 'search_indexer') else None
        converters = self.Model.Meta.search_converters if hasattr(self.Model.Meta, 'search_converters') else None
        callback = self.Model.Meta.search_callback if hasattr(self.Model.Meta, 'search_callback') else None
        index_entity(
            instance=instance,
            index=index_for(self.Model),
            only=only,
            exclude=exclude,
            indexer=indexer,
            extra_converters=converters,
            callback=callback)

    def before_delete(self, key):
        unindex_entity(key, index_for(self.Model))


def index_for(Model):
    if hasattr(Model.Meta, 'search_index'):
        return Model.Meta.search_index
    else:
        return 'searchable:%s' % Model._get_kind()


def default_entity_indexer(instance, properties, extra_converters=None):
    results = []

    converters = {}
    converters.update(property_to_field_map)
    if extra_converters:
        converters.update(extra_converters)

    for property in properties:
        value = getattr(instance, property)
        converted = None
        property_instance = instance._properties[property]
        property_class = property_instance.__class__
        converter = converters.get(property_class, converters.get(property, None))

        if not value or not converter:
            if property_class in (ndb.KeyProperty, ndb.BlobKeyProperty):
                logging.debug("Search utilities will not automatically index Key or BlobKey property %s" % property)
            continue

        if not property_instance._repeated:
            converted = converter(property, value)
        else:
            if property_class not in non_repeatable_properties:
                converted = [converter(property, x) for n, x in enumerate(value)]
            else:
                logging.debug("Could not automatically add field %s to the index because date and number fields can not be repeated." % property)

        if not converted:
            continue

        if isinstance(converted, (list, tuple)):
            results.extend(converted)
        else:
            results.append(converted)

    return results


def index_entity(instance, index, only=None, exclude=None, extra_converters=None, indexer=None, callback=None):
    """
    Adds an Model instance into full-text search indexes.

    :param instance: an instance of ndb.Model
    :param list(string) only: If provided, will only index these fields
    :param list(string) exclude: If provided, will not index any of these fields
    :param dict extra_converters: Extra map of property names or types to converter functions.
    :param indexer: A function that transforms properties into search index fields.
    :param callback: A function that will recieve (instance, fields).
        Fields is a map of property names to search. Field instances generated by the indexer
        the callback can modify this dictionary to change how the item is indexed.

    This is usually done in :meth:`Model.after_put <ferris3.ndb.Model.after_put>`, for example::

        def after_put(self):
            index(self)

    """

    indexer = indexer if indexer else default_entity_indexer
    indexes = index if isinstance(index, (list, tuple)) else [index]
    only = only if only else [k for k in instance._properties.keys() if hasattr(instance, k)]
    exclude = exclude if exclude else []
    properties = [x for x in only if x not in exclude]

    fields = indexer(instance, properties, extra_converters=extra_converters)

    if callback:
        callback(instance=instance, fields=fields)

    try:
        doc = search_api.Document(doc_id=str(instance.key.urlsafe()), fields=fields)

        for index_name in indexes:
            index = search_api.Index(name=index_name)
            index.put(doc)

    except Exception as e:
        logging.error("Adding model %s instance %s to the full-text index failed" % (instance.key.kind(), instance.key.id()))
        logging.error("Search API error: %s" % e)
        logging.debug([(x.name, x.value) for x in fields])


def unindex_entity(instance_or_key, index=None):
    """
    Removes a document from the full-text search.

    This is usually done in :meth:`Model.after_delete <ferris3.ndb.Model.after_delete>`, for example::

        @classmethod
        def after_delete(cls, key):
            unindex(key)

    """
    if isinstance(instance_or_key, ndb.Model):
        instance_or_key = instance_or_key.key

    indexes = index if isinstance(index, (list, tuple)) else [index]

    for index_name in indexes:
        index = search_api.Index(name=index_name)
        index.delete(str(instance_or_key.urlsafe()))


def to_entities(results):
    """
    Transform a list of search results into ndb.Model entities by using the document id
    as the urlsafe form of the key.
    """
    if isinstance(results, SearchResults):
        items = [x for x in ndb.get_multi([ndb.Key(urlsafe=y.doc_id) for y in results.items]) if x]
        return SearchResults(items=items, error=results.error, next_page_token=results.next_page_token)
    else:
        return[x for x in ndb.get_multi([ndb.Key(urlsafe=y.doc_id) for y in results]) if x]


SearchResults = namedtuple('SearchResults', ['items', 'error', 'next_page_token'])


def search(index, query, sort=None, sort_default_values=None, limit=None, page_token=None, ids_only=True, options=None, per_document_cursor=False):
    """
    Searches an index with the given query and returns a list of document ids or search documents.

     To get the full search document pass ``ids_only = False``.

    :param index: The name of the index to search.
    :param query: Query string as described in `the App Engine documentation <https://developers.google.com/appengine/docs/python/search/query_strings>`__.
    :param sort: A sort string, can be ``"field_name"`` for ascending or ``"-field_name"`` for descending. Can also be a list of sorts, such as ``["price", "-rating"]``.
    :param sort_default_values: The default value to use for sorting if there is no value in the document.
    :param limit: Maximum number of results to return.
    :param page_token: Cursor used to get a particular page of results.
    :param ids_only: By default, this only returns document ids as the most common use case is to get the entity or
        database entries associated with the document. Pass ``False`` here to get the complete document.
    :param options: Advanced options that are passed directly to ``index.search``. See `query options <https://developers.google.com/appengine/docs/python/search/options>`__.
    :param per_document_cursor: Whether to include a cursor for every document or not.
    :returns: a tuple of ``(items, error, next_page_token)``

    Examples::

        # Search all pages for "policies"
        results, error, next_page_token = search('searchable:Page', 'policies', limit=20)

        # Search all products for "rake" sorted by price descending
        results, error, next_page_token = search('searchable:Page', 'rake', sort='-price', limit=20)
    """

    options = options if options else {}
    error = None
    results = []
    current_cursor = None
    next_cursor = None

    try:
        index = search_api.Index(name=index)
        current_cursor = search_api.Cursor(web_safe_string=page_token) if page_token else search_api.Cursor(per_result=per_document_cursor)

        options_params = dict(
            limit=limit,
            ids_only=ids_only,
            cursor=current_cursor)

        if sort:
            options_params['sort_options'] = create_sort_options(sort, sort_default_values)

        options_params.update(options)

        # if limit is none, remove it, as it'll cause issues.
        if options_params.get('limit') is None:
            del options_params['limit']

        query = search_api.Query(query_string=query, options=search_api.QueryOptions(**options_params))
        index_results = index.search(query)

        results = list(index_results)

        current_cursor = current_cursor.web_safe_string if current_cursor else None
        next_cursor = index_results.cursor.web_safe_string if index_results.cursor and results else None

    except (search_api.Error, search_api.query_parser.QueryException) as e:
        error = str(e)

    return SearchResults(items=results, error=error, next_page_token=next_cursor)


def create_sort_options(fields, default_values=None):
    default_values = default_values or {}
    expressions = []

    fields = fields if isinstance(fields, (list, tuple)) else [fields]

    for field in fields:
        if isinstance(field, search_api.SortExpression):
            expressions.append(field)
            continue

        if field.startswith('-'):
            field = field[1:]
            direction_exp = search_api.SortExpression.DESCENDING
        else:
            direction_exp = search_api.SortExpression.ASCENDING

        default_value = default_values.get(field, '')

        if inspect.isfunction(default_value):
            default_value = default_value(field, direction_exp)

        expressions.append(
            search_api.SortExpression(
                expression=field,
                direction=direction_exp,
                default_value=default_value
            ))

    return search_api.SortOptions(expressions=expressions)


def join_query(filters, operator='AND', parenthesis=False):
    """
    Utility function for joining muliple queries together
    """
    operator = ' %s ' % operator
    filters = [x for x in filters if x]
    if parenthesis:
        filters = ["(%s)" % x for x in filters]
    return operator.join(filters)



import functools
import logging
import types
import inspect


def partial(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        return functools.partial(func, *args, **kwargs)
    setattr(inner, '_f3_partial', True)
    return inner


def tap(func, data):
    func(data)
    return data


def pipe(func, data):
    return func(data)


def raise_if(value_or_func, ex, data):
    if callable(value_or_func):
        if value_or_func(data):
            raise ex
    elif value_or_func == data:
        raise ex

    return data


class mixedmethod(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        return functools.partial(self.func, instance, cls)


class Chain(object):
    def __init__(self, value=None, use=None):
        self._modules = {}
        self.set_value(value)

        if use:
            for m in use:
                self.add_chain_module(m)

    def _chain_ref(self):
        return self

    def get_value(self):
        return self._value

    def set_value(self, value):
        self._value = value
        return self

    value = get_value

    @mixedmethod
    def add_chain_function(self, cls, func, name=None):
        if not name:
            name = func.__name__

        # Curry if needed
        if not hasattr(func, '_f3_partial'):
            func = partial(func)

        # Wrap with the call adapter
        @functools.wraps(func)
        def call_wrapper(func):
            def inner(self, *args, **kwargs):
                val = self.get_value()
                try:
                    val = func(*args, **kwargs)(val)
                except:
                    logging.error("Problem occurred while executing chain '%s' with data '%s' and args %s, %s" % (func.__name__, val, args, kwargs))
                    raise

                self.set_value(val)
                return self._chain_ref()
            return inner

        func = call_wrapper(func)

        if self:
            setattr(self, name, types.MethodType(func, self))
        else:
            setattr(cls, name, func)

    @classmethod
    def add_chain_functions(cls, *funcs):
        for func in funcs:
            cls.add_chain_function(func)

    @mixedmethod
    def add_chain_module(self, cls, module, module_name=None, only=None, exclude=None):
        if not module_name:
            module_name = module.__name__.split('.').pop()

        def get_mod(self):
            if module_name in self._modules:
                return self._modules[module_name]

            cm = ChainModule(self)
            for fname, func in inspect.getmembers(module, callable):
                if only and fname not in only:
                    continue
                if exclude and fname in exclude:
                    continue
                cm.add_chain_function(func, fname)

            self._modules[module_name] = cm
            return cm

        # If it's a class, add it to all instances and lazy load
        if not self:
            setattr(cls, module_name, property(get_mod))

        # If this is an instance, just go ahead and add it.
        else:
            setattr(self, module_name, get_mod(self))


class ChainModule(Chain):
    def __init__(self, parent):
        self._parent = parent

    def _chain_ref(self):
        return self._parent

    def get_value(self):
        return self._parent.get_value()

    def set_value(self, value):
        self._parent.set_value(value)
        return self._parent

    value = get_value

    add_chain_module = None


# Add built-ins
Chain.add_chain_functions(
    tap,
    pipe,
    raise_if
)



import ferris3

#
# Method implementations.
# Bare-bones, can be re-used in other methods and such.
#


def list_impl(ListMessage, query):
    return ferris3.ToolChain(query) \
        .messages.serialize_list(ListMessage) \
        .value()


def paginated_list_impl(ListMessage, query, limit, page_token):
    return ferris3.ToolChain(query) \
        .ndb.paginate(limit=limit, page_token=page_token) \
        .messages.serialize_list(ListMessage) \
        .value()


def searchable_list_impl(ListMessage, index, query, limit, sort, page_token):
    def check_for_search_errors(data):
        if data.error:
            raise ferris3.BadRequestException("Search error: %s" % data.error)

    return ferris3.ToolChain(query) \
        .search.search(index, sort=sort, limit=limit, page_token=page_token) \
        .tap(check_for_search_errors) \
        .search.to_entities() \
        .messages.serialize_list(ListMessage) \
        .value()


def get_impl(Model, Message, item_key):
    return ferris3.ToolChain(item_key) \
        .ndb.get() \
        .raise_if(None, ferris3.NotFoundException()) \
        .ndb.check_kind(Model) \
        .messages.serialize(Message) \
        .value()


def delete_impl(Model, item_key):
    return ferris3.ToolChain(item_key) \
        .ndb.key() \
        .ndb.check_kind(Model) \
        .ndb.delete() \
        .value()


def update_impl(Model, Message, item_key, request):
    item = ferris3.ToolChain(item_key) \
        .ndb.get() \
        .raise_if(None, ferris3.NotFoundException()) \
        .ndb.check_kind(Model) \
        .value()

    return ferris3.ToolChain(request) \
        .messages.deserialize(item) \
        .ndb.put() \
        .messages.serialize(Message) \
        .value()


def insert_impl(Model, Message, request):
    return ferris3.ToolChain(request) \
        .messages.deserialize(Model) \
        .ndb.put() \
        .messages.serialize(Message) \
        .value()

#
# Full method wrappers.
# Can be used as endpoint methods directly.
#


def list(Model, Message=None, ListMessage=None, query=None, name='list'):
    """
    Implements a very simple list method for the given model.

    .. warning:: No limiting logic is applied to the query so this will attempt to get all results.
        It is almost always preferable to use :func:`paginated_list`.

    """
    if not Message:
        Message = ferris3.model_message(Model)

    if not ListMessage:
        ListMessage = ferris3.list_message(Message)

    if not query:
        query = Model.query()

    if callable(query):
        query = query()

    @ferris3.auto_method(returns=ListMessage, name=name)
    def inner(self, request):
        return list_impl(ListMessage, query)

    return inner


def paginated_list(Model, Message=None, ListMessage=None, query=None, limit=50, name='paginated_list'):
    """
    Similar to :func:`list` but returns results in pages of ``limit`` size.
    """
    if not Message:
        Message = ferris3.model_message(Model)

    if not ListMessage:
        ListMessage = ferris3.list_message(Message)

    if not query:
        query = Model.query()

    if callable(query):
        query = query()

    @ferris3.auto_method(returns=ListMessage, name=name)
    def inner(self, request, page_token=(str, '')):
        return paginated_list_impl(ListMessage, query, limit, page_token)

    return inner


def searchable_list(Model=None, Message=None, ListMessage=None, limit=50, index=None, name='search'):
    """
    Implements a search method by using the :mod:`~ferris3.search` module. This method assumes you are
    using the common use case of indexing datastore entities and does not work for generic searches.
    """
    if not Message:
        Message = ferris3.model_message(Model)

    if not ListMessage:
        ListMessage = ferris3.list_message(Message)

    if not index:
        index = ferris3.search.index_for(Model)

    @ferris3.auto_method(returns=ListMessage, name=name)
    def inner(self, request, query=(str, ''), sort=(str, None), page_token=(str, '')):
        return searchable_list_impl(ListMessage, index, query, limit, sort, page_token)

    return inner


def get(Model, Message=None, name='get'):
    """
    Implements a straightfoward get method by using the urlsafe version of the entity's key.
    """
    if not Message:
        Message = ferris3.model_message(Model)

    @ferris3.auto_method(returns=Message, name=name)
    def inner(self, request, item_key=(str,)):
        return get_impl(Model, Message, item_key)

    return inner


def delete(Model, name='delete'):
    """
    Implements a straightfoward delete method by using the urlsafe version of the entity's key.
    """

    @ferris3.auto_method(name=name, http_method='DELETE')
    def inner(self, request, item_key=(str,)):
        delete_impl(Model, item_key)
        return None

    return inner


def insert(Model, Message=None, name='insert'):
    """
    Implements the insert method. The request fields are determined by the ``Message`` parameter.
    """
    if not Message:
        Message = ferris3.model_message(Model)

    @ferris3.auto_method(returns=Message, name=name, http_method='POST')
    def inner(self, request=(Message,)):
        return insert_impl(Model, Message, request)

    return inner


def update(Model, Message=None, name='update'):
    """
    Implements the update method. The item updated is determined by the urlsafe key of that item.
    The request fields are determined by the ``Message`` parameter.
    """
    if not Message:
        Message = ferris3.model_message(Model)

    @ferris3.auto_method(returns=Message, name=name, http_method='POST')
    def inner(self, request=(Message,), item_key=(str,)):
        return update_impl(Model, Message, item_key, request)

    return inner



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from google.appengine.api import memcache
from google.appengine.ext import ndb
from functools import wraps
import datetime
import threading
import inspect


none_sentinel_string = u'☃☸☃ - caching sentinel'


def cache(key, ttl=0, backend=None):
    """
    General-purpose caching decorator. This decorator causes the result of a function
    to be cached so that subsequent calls will return the cached result instead of
    calling the function again. The ttl argument determines how long the cache is valid,
    once the cache is invalid the function will be called to generate a new value and the
    cache will be refreshed. The backend argument can be used to determine how the value
    is cached- by default, the value is stored in memcache but there are built-in backends
    for thread-local caching and caching via the datastore.

    Example::

        @cache('something_expensive', ttl=3600)
        def expensive_function():
            ...

    """
    if backend is None or backend == 'memcache':
        backend = MemcacheBackend
    elif backend == 'local':
        backend = LocalBackend
    elif backend == 'datastore':
        backend = DatastoreBackend

    def wrapper(f):
        @wraps(f)
        def dispatcher(*args, **kwargs):
            data = backend.get(key)

            if data == none_sentinel_string:
                return None

            if data is None:
                data = f(*args, **kwargs)
                backend.set(key, none_sentinel_string if data is None else data, ttl)

            return data

        def cache_getter():
            data = backend.get(key)
            if data == none_sentinel_string:
                return None
            return data

        setattr(dispatcher, 'clear_cache', lambda: backend.delete(key))
        setattr(dispatcher, 'cached', cache_getter)
        setattr(dispatcher, 'uncached', f)
        return dispatcher
    return wrapper


def cache_by_args(key, ttl=0, backend=None):
    """
    Like :func:`cache`, but will use any arguments to the function as part of the key to
    ensure that variadic functions are cached separately. Argument must be able to be
    printed as a string- it's recommended to use plain data types as arguments.
    """
    def wrapper(f):
        argspec = inspect.getargspec(f)[0]

        if len(argspec) and argspec[0] in ('self', 'cls'):
            is_method = True
        else:
            is_method = False

        @wraps(f)
        def dispatcher(*args, **kwargs):
            targs = args if not is_method else args[1:]
            arg_key = "%s:%s:%s" % (key, targs, kwargs)

            @cache(arg_key, ttl, backend=backend)
            def inner_dispatcher():
                return f(*args, **kwargs)

            return inner_dispatcher()
        return dispatcher
    return wrapper


def cache_using_local(key, ttl=0):
    """
    Shortcut decorator for caching using the thread-local cache.
    """
    return cache(key, ttl, backend=LocalBackend)


def cache_using_memcache(key, ttl=0):
    """
    Shortcut decorator for caching using the memcache.
    """
    return cache(key, ttl, backend=MemcacheBackend)


def cache_using_datastore(key, ttl=0):
    """
    Shortcut decorator for caching using the datastore
    """
    return cache(key, ttl, backend=DatastoreBackend)


def cache_by_args_using_local(key, ttl=0):
    """
    Shortcut decorator for caching by arguments using the thread-local cache.
    """
    return cache_by_args(key, ttl, backend=LocalBackend)


def cache_by_args_using_memcache(key, ttl=0):
    """
    Shortcut decorator for caching by arguments using the memcache.
    """
    return cache_by_args(key, ttl, backend=MemcacheBackend)


def cache_by_args_using_datastore(key, ttl=0):
    """
    Shortcut decorator for caching by arguments using the datastore
    """
    return cache_by_args(key, ttl, backend=DatastoreBackend)


class LocalBackend(object):
    """
    The local backend stores caches in a thread-local variable. The caches are available
    for this thread and likely just for the duration of one request.
    """
    cache_obj = threading.local()

    @classmethod
    def set(cls, key, data, ttl):
        if ttl:
            expires = datetime.datetime.now() + datetime.timedelta(seconds=ttl)
        else:
            expires = None
        setattr(cls.cache_obj, key, (data, expires))

    @classmethod
    def get(cls, key):
        if not hasattr(cls.cache_obj, key):
            return None

        data, expires = getattr(cls.cache_obj, key)

        if expires and expires < datetime.datetime.now():
            delattr(cls.cache_obj, key)
            return None

        return data

    @classmethod
    def delete(cls, key):
        try:
            delattr(cls.cache_obj, key)
        except AttributeError:
            pass

    @classmethod
    def reset(cls):
        for a in cls.cache_obj.__dict__.keys():
            delattr(cls.cache_obj, a)


class MemcacheBackend(object):
    """
    Stores caches in memcache. Memcache is available across instances but is subject to
    being dumped from the cache before the expiration time.
    """
    @classmethod
    def set(cls, key, data, ttl):
        memcache.set(key, data, ttl)

    @classmethod
    def get(cls, key):
        return memcache.get(key)

    @classmethod
    def delete(cls, key):
        memcache.delete(key)


class MemcacheCompareAndSetBackend(MemcacheBackend):
    """
    Same as the regular memcache backend but uses compare-and-set logic to ensure
    that memcache updates are atomic.
    """
    @classmethod
    def set(cls, key, data, ttl):
        client = memcache.Client()
        if not client.gets(key):
            memcache.set(key, data, ttl)
            return

        for _ in range(10):
            if client.cas(key, data, ttl):
                break


class DatastoreBackend(object):
    """
    Stores caches in the datastore which has the effect of them being durable and persistent,
    unlike the memcache and local backends. Items stored in the datastore are certain to remain
    until the expiration time passes.
    """
    @classmethod
    def set(cls, key, data, ttl):
        if ttl:
            expires = datetime.datetime.now() + datetime.timedelta(seconds=ttl)
        else:
            expires = None

        DatastoreCacheModel(id=key, data=data, expires=expires).put()

    @classmethod
    def get(cls, key):
        item = ndb.Key(DatastoreCacheModel, key).get()

        if not item:
            return None

        if item.expires and item.expires < datetime.datetime.now():
            item.key.delete()
            return None

        return item.data

    @classmethod
    def delete(cls, key):
        ndb.Key(DatastoreCacheModel, key).delete()


class DatastoreCacheModel(ndb.Model):
    data = ndb.PickleProperty(indexed=False, compressed=True)
    expires = ndb.DateTimeProperty(indexed=False)


class LayeredBackend(object):
    """
    Allows you to use multiple backends at once. When an item is cached it is put
    in to each backend. Retrieval checks each backend in order for the item. This is
    very useful when combining fast but volatile backends (like local) with slow
    but durable backends (like datastore).

    Example::

        @cache('something_expensive', ttl=3600, backend=LayeredBackend(LocalBackend, DatastoreBackend))
        def expensive_function():
            ...

    """
    def __init__(self, *args):
        self.backends = args

    def set(self, key, data, ttl):
        for b in self.backends:
            b.set(key, data, ttl)

    def get(self, key):
        for b in self.backends:
            data = b.get(key)
            if data is not None:
                return data

    def delete(self, key):
        for b in self.backends:
            b.delete(key)



"""
Utilities that build on top of the built-in ndb.
"""

from google.appengine.ext import ndb
from google.appengine.datastore.datastore_query import Cursor
import types
from collections import namedtuple


class ModelMeta(ndb.model.MetaModel):
    """
    Ensures behaviors are constructed.
    """
    def __init__(cls, name, bases, dct):
        super(ModelMeta, cls).__init__(name, bases, dct)

        # Make sure the Meta class has a proper chain
        if cls.__name__ != 'Model' and not issubclass(cls.Meta, Model.Meta):
            cls.Meta = type('Meta', (cls.Meta, Model.Meta), {})

        # Behaviors
        setattr(cls, 'behaviors', [x(cls) for x in cls.Meta.behaviors])


class Model(ndb.Model):
    """
    ``ndb.Model`` subclass that supports easier callbacks and behaviors.
    """
    __metaclass__ = ModelMeta

    class Meta(object):
        behaviors = ()

    def before_put(self):
        """
        Called before an item is saved.

        :arg self: refers to the item that is about to be saved
        :note: ``self.key`` is invalid if the current item has never been saved
        """
        pass

    def after_put(self, key):
        """
        Called after an item has been saved.

        :arg self: refers to the item that has been saved
        :arg key: refers to the key that the item was saved as
        """
        pass

    @classmethod
    def before_delete(cls, key):
        """
        Called before an item is deleted.

        :arg key: is the key of the item that is about to be deleted. It is okay to ``get()`` this key to interogate the properties of the item.
        """
        pass

    @classmethod
    def after_delete(cls, key):
        """
        Called after an item is deleted.

        :arg key: is the key of the item that was deleted. It is not possible to call ``get()`` on this key.
        """
        pass

    @classmethod
    def before_get(cls, key):
        """
        Called before an item is retrieved. Note that this does not occur for queries.

        :arg key: Is the key of the item that is to be retrieved.
        """
        pass

    @classmethod
    def after_get(cls, key, item):
        """
        Called after an item has been retrieved. Note that this does not occur for queries.

        :arg key: Is the key of the item that was retrieved.
        :arg item: Is the item itself.
        """
        pass

    # Impl details

    @classmethod
    def _invoke_behaviors(cls, method, *args, **kwargs):
        for b in cls.behaviors:
            getattr(b, method)(*args, **kwargs)

    def _pre_put_hook(self):
        self._invoke_behaviors('before_put', self)
        return self.before_put()

    def _post_put_hook(self, future):
        res = future.get_result()
        self._invoke_behaviors('after_put', self)
        return self.after_put(res)

    @classmethod
    def _pre_delete_hook(cls, key):
        cls._invoke_behaviors('before_delete', key)
        return cls.before_delete(key)

    @classmethod
    def _post_delete_hook(cls, key, future):
        cls._invoke_behaviors('after_delete', key)
        return cls.after_delete(key)

    @classmethod
    def _pre_get_hook(cls, key):
        cls._invoke_behaviors('before_get', key)
        return cls.before_get(key)

    @classmethod
    def _post_get_hook(cls, key, future):
        res = future.get_result()
        cls._invoke_behaviors('after_get', res)
        return cls.after_get(key, res)


class Behavior(object):
    """
    Behaviors allow you to encapsulate callbacks into discrete units that can be combined together.
    """
    def __init__(self, Model):
        self.Model = Model

    def before_put(self, instance):
        pass

    def after_put(self, instance):
        pass

    def before_delete(self, key):
        pass

    def after_delete(self, key):
        pass

    def before_get(self, key):
        pass

    def after_get(self, item):
        pass


def key(s):
    if isinstance(s, ndb.Model):
        return s.key
    if isinstance(s, ndb.Key):
        return s
    if isinstance(s, types.StringTypes):
        return ndb.Key(urlsafe=s)
    return None


def get(item):
    if isinstance(item, ndb.Query):
        return item.get()
    return key(item).get()


def put(item):
    item.put()
    return item


def delete(item):
    key(item).delete()
    return item


def check_kind(kind, item):
    if issubclass(kind, ndb.Model):
        kind = kind._get_kind()

    item_key = key(item)
    if not kind == item_key.kind():
        raise ValueError("Incorrect kind %s, expected %s" % (item_key.kind(), kind))

    return item


PaginationResults = namedtuple('PaginationResults', ['items', 'next_page_token'])


def paginate(query, limit=50, page_token=None):
    if page_token and not isinstance(page_token, Cursor):
        page_token = Cursor(urlsafe=page_token)

    # Force all falsy values into None
    if not page_token:
        page_token = None

    data, next_cursor, more = query.fetch_page(limit, start_cursor=page_token)

    return PaginationResults(items=data, next_page_token=next_cursor.urlsafe() if more else None)



import protopigeon


def serialize(MessageType, entity):
    return protopigeon.to_message(entity, MessageType)


def deserialize(Model, message):
    return protopigeon.to_entity(message, Model)


def serialize_list(ListMessageType, entities):
    from .ndb import PaginationResults
    from .search import SearchResults

    if isinstance(entities, (PaginationResults, SearchResults)):
        next_page_token = entities.next_page_token
        entities = entities.items
    else:
        next_page_token = None

    MessageType = ListMessageType.items.message_type

    message = ListMessageType()
    message.items = [serialize(MessageType, x) for x in entities]
    message.next_page_token = next_page_token

    return message



from __future__ import absolute_import
import yaml
import os
import endpoints

_apis = {}
_default_api_name = None

base_directory = os.getcwd()


def add(config_file, default=False):
    global _apis, _default_api_name

    config = load_config_file(config_file)
    api = endpoints.api(**config)
    _apis[config['name']] = api

    if default:
        _default_api_name = config['name']

    return api


def get(name=None):
    if not name:
        name = _default_api_name
    return _apis.get(name)


def default():
    return get()


def get_all():
    return _apis.values()


def load_config_file(config_file):
    with open(os.path.join(base_directory, config_file)) as f:
        config = yaml.load(f)

    # Replace constants
    recursive_replace(config, 'API_EXPLORER_CLIENT_ID', endpoints.API_EXPLORER_CLIENT_ID)
    recursive_replace(config, 'USERINFO', 'https://www.googleapis.com/auth/userinfo.email')

    auth_level = config.get('auth_level', 'optional')
    config['auth_level'] = {
        'required': endpoints.AUTH_LEVEL.REQUIRED,
        'optional': endpoints.AUTH_LEVEL.OPTIONAL,
        'optional_continue': endpoints.AUTH_LEVEL.OPTIONAL_CONTINUE,
        'none': endpoints.AUTH_LEVEL.NONE
    }.get(auth_level)

    return config


def recursive_replace(container, old, new):
    if isinstance(container, dict):
        for key, value in container.iteritems():
            if value == old:
                container[key] = new
            if isinstance(value, (dict, list)):
                recursive_replace(value, old, new)
    else:
        for ix, value in enumerate(container):
            if value == old:
                container[ix] = new



from __future__ import absolute_import
from google.appengine.ext import ndb
import hashlib
import endpoints
import os
from oauth2client.client import AccessTokenCredentials


def get_endpoints_credentials():
    """
    Gets the oauth2 credentials from the user authenticated to Google Cloud Endpoints.

    Presently, this does not work for Android and iOS clients. We are open to patches to fix it.
    """
    user = endpoints.get_current_user()
    if not user:
        return False

    if not 'HTTP_AUTHORIZATION' in os.environ:
        return False

    token = os.environ['HTTP_AUTHORIZATION'].split(' ').pop()
    credentials = AccessTokenCredentials(token, 'appengine:ferris')

    return credentials


def _get_config():
    from . import settings
    config = settings.get('oauth2_service_account')
    if not config:
        raise RuntimeError("OAuth2 Service Account is not configured correctly")
    return config


try:
    from oauth2client.client import SignedJwtAssertionCredentials
except ImportError:
    SignedJwtAssertionCredentials = None

from oauth2client.appengine import StorageByKeyName, CredentialsNDBProperty


def build_service_account_credentials(scope, user=None):
    """
    Builds service account credentials using the configuration stored in :mod:`~ferris3.settings`
    and masquerading as the provided user.
    """
    if not SignedJwtAssertionCredentials:
        raise EnvironmentError("Service account can not be used because PyCrypto is not available. Please install PyCrypto.")

    config = _get_config()

    if not isinstance(scope, (list, tuple)):
        scope = [scope]

    key = _generate_storage_key(config['client_email'], scope, user)
    storage = StorageByKeyName(ServiceAccountStorage, key, 'credentials')

    creds = SignedJwtAssertionCredentials(
        service_account_name=config['client_email'],
        private_key=config['private_key'],
        scope=scope,
        prn=user)
    creds.set_store(storage)

    return creds


class ServiceAccountStorage(ndb.Model):
    """
    Tracks access tokens in the database. The key is
    based on the scopes, user, and clientid
    """
    credentials = CredentialsNDBProperty()

    @classmethod
    def _get_kind(cls):
        return '_ferris_OAuth2ServiceAccountStorage'


def _generate_storage_key(client_id, scopes, user):
    s = u"%s%s%s" % (client_id, sorted(scopes), user)
    hash = hashlib.sha1(s.encode())
    return hash.hexdigest()



import jinja2
import os

_debug = os.environ.get('SERVER_SOFTWARE', '').startswith('Dev')
_environment = None


def environment():
    global _environment
    if _environment:
        return _environment

    jinja2_env_kwargs = {
        'loader': jinja2.FileSystemLoader(os.getcwd()),
        'auto_reload': False,
        'cache_size': 0 if _debug else 50,
    }
    _environment = jinja2.Environment(**jinja2_env_kwargs)
    return _environment


def render(name, context=None):
    env = environment()
    template = env.get_or_select_template(name)
    context = context if context else {}

    result = template.render(context)
    return result



from .chain import Chain
from . import ndb, messages, search


class ToolChain(Chain):
    pass


ToolChain.add_chain_module(ndb)
ToolChain.add_chain_module(messages)
ToolChain.add_chain_module(search, only=("search", "to_entities"))



from __future__ import absolute_import
import logging

_defaults = {}
_sentinel = object()


class ConfigurationError(Exception):
    pass


def load(refresh=False):
    """
    Executed when the project is created and loads the settings from app/settings.py
    """
    global _defaults

    if _defaults and not refresh:
        return

    try:
        import app.settings as appsettings
        reload(appsettings)
    except ImportError:
        raise ConfigurationError("Settings not found. Please create app/settings.py")

    try:
        appdefaults = appsettings.settings
    except AttributeError:
        raise ConfigurationError("No dictionary 'settings' found in settings.py")

    logging.info("Ferris settings loaded")

    defaults(appdefaults)


def defaults(dict=None):
    """
    Adds a set of default values to the settings registry. These can and will be updated
    by any settings modules in effect, such as the Settings Manager.

    If dict is None, it'll return the current defaults.
    """
    if dict:
        _defaults.update(dict)
    else:
        return _defaults


def all():
    """
    Returns the entire settings registry
    """
    settings = {}
    settings.update(_defaults)
    return settings


def set(key, value):
    global _defaults
    _defaults[key] = value


def get(key, default=_sentinel):
    """
    Returns the setting at key, if available, raises an ConfigurationError if default is none, otherwise
    returns the default
    """
    _settings = all()
    if key not in _settings:
        if default is _sentinel:
            raise ConfigurationError("Missing setting %s" % key)
        else:
            return default
    return _settings[key]



import os
import importlib
import inspect
from protorpc import remote


base_directory = os.path.abspath(os.getcwd())
app_directory = os.path.join(base_directory, 'app')


def discover_api_services():
    api_module_files = discover_files(lambda x: x.endswith('service.py'))
    modules = load_modules_from_files(api_module_files)
    apis = find_service_classes(modules)
    return [x[1] for x in apis]


def discover_webapp2_routes():
    handler_module_files = discover_files(lambda x: x.endswith('handler.py'))
    modules = load_modules_from_files(handler_module_files)
    routes = find_webapp2_routes(modules)
    return routes


def discover_files(predicate):
    results = []
    for root_path, _, files in os.walk(app_directory):
        for filename in files:
            if predicate(filename):
                results.append(os.path.join(root_path, filename))
    return results


def load_modules_from_files(files):
    modules = []
    for filename in files:
        # /home/.../app/thing/thing_service.py -> app/thing/thing_service.py
        relative_path = os.path.relpath(filename, base_directory)
        # app/thing/thing_service.py -> app/thing/thing_service
        module_path, ext = os.path.splitext(relative_path)
        # app/thing/thing_service -> app.thing.thing_service
        import_path = '.'.join(module_path.split(os.sep))

        module = importlib.import_module(import_path)
        modules.append(module)
    return modules


def find_service_classes(modules):
    classes = []
    for module in modules:
        apis = inspect.getmembers(module, lambda x: inspect.isclass(x) and is_remote_service(x))
        classes.extend(apis)
    return classes


def find_webapp2_routes(modules):
    routes = []
    for module in modules:
        if hasattr(module, 'webapp2_routes'):
            routes.extend(module.webapp2_routes)
    return routes


def is_remote_service(cls):
    # This is kind of an ugly hack, for some reason issubclass doesn't always work.
    # Only include items with api_info, this prevents the baseclass from being included
    if not hasattr(cls, 'api_info'):
        return False

    # check up the mro chain for inheritence.
    for mro in inspect.getmro(cls):
        if mro.__module__ == remote.Service.__module__ and mro.__name__ == remote.Service.__name__:
            return True
    return False



from __future__ import absolute_import
from protorpc import message_types
from protorpc import messages
import endpoints
from .anodi import annotated
import inspect
import re
import yaml
import os

_endpoints = {}
_default_endpoint_name = None

base_directory = os.getcwd()


def add(config_or_file, default=False):
    """
    Add an endpoint to the registry.

    ``config_or_file`` can be the path to a yaml definition file or a dictionary of arguments to pass to
    ``endpoints.api``. See also Google's documentation on `endpoints.api <https://developers.google.com/appengine/docs/python/endpoints/create_api#defining_the_api_endpointsapi>`__.

    Tpyically, this is called in an application's ``main.py`` before any services are loaded.

    Examples::

        ferris3.endpoints.add('app/default-endpoint.yaml', default=True)
        ferris3.endpoints.add({
            name: 'test',
            version: 'v1'
        })
    """
    global _endpoints, _default_endpoint_name

    if isinstance(config_or_file, (str, unicode)):
        config = load_config_file(config_or_file)
    else:
        config = config_or_file

    api = endpoints.api(**config)
    _endpoints[config['name']] = api

    if default:
        _default_endpoint_name = config['name']

    return api


def get(name=None):
    """
    Get an endpoint by name from the registry.

    The value returned is a normal `endpoints.api <https://developers.google.com/appengine/docs/python/endpoints/create_api#creating_an_api_implemented_with_multiple_classes>`_ class that can be used exactly as descibed in the Google documentionat.

    ``name`` is the value of the ``name`` configuration field given for the endpoint. If no name is provided, it'll return the default endpoint.

    Examples::

        endpoint = ferris3.endpoints.get('ferris')

        @endpoint.api_class(resource_name='shelves')
        class ShelvesService(remote.Service):
            ...

    """
    if not name:
        if not _default_endpoint_name:
            raise RuntimeError("No default endpoint has been configured")
        name = _default_endpoint_name

    return _endpoints.get(name)


def default():
    """
    Simply calls :func:`get` for the default endpoint.
    """
    return get()


def get_all():
    return _endpoints.values()


def load_config_file(config_file):
    with open(os.path.join(base_directory, config_file)) as f:
        config = yaml.load(f)

    # Replace constants
    recursive_replace(config, 'API_EXPLORER_CLIENT_ID', endpoints.API_EXPLORER_CLIENT_ID)
    recursive_replace(config, 'USERINFO', 'https://www.googleapis.com/auth/userinfo.email')

    auth_level = config.get('auth_level', 'optional')
    config['auth_level'] = {
        'required': endpoints.AUTH_LEVEL.REQUIRED,
        'optional': endpoints.AUTH_LEVEL.OPTIONAL,
        'optional_continue': endpoints.AUTH_LEVEL.OPTIONAL_CONTINUE,
        'none': endpoints.AUTH_LEVEL.NONE
    }.get(auth_level)

    return config


def recursive_replace(container, old, new):
    if isinstance(container, dict):
        for key, value in container.iteritems():
            if value == old:
                container[key] = new
            if isinstance(value, (dict, list)):
                recursive_replace(value, old, new)
    else:
        for ix, value in enumerate(container):
            if value == old:
                container[ix] = new


def underscore(string):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def auto_service(cls=None, endpoint=None, **kwargs):
    """
    Automatically configures and adds a ``remote.Service`` class to the endpoint service registry.

    If ``endpoint`` is None then the default endpoint will be used. If it is a string then the endpoint will be looked up in the registry using :func:`get`.

    The ``resource_name`` and ``path`` configuration options are automatically determined from the class name.
    For example, "PostsService" becomes "posts" and "FuzzyBearsService" becomes "fuzzy_bears". These can be overriden by
    passing in the respecitve kwargs.

    All kwargs are passed directly through to `endpoints.api_class <https://developers.google.com/appengine/docs/python/endpoints/create_api#creating_an_api_implemented_with_multiple_classes>`_.

    Examples::

        @ferris3.auto_service
        class PostsService(ferris3.Service):
            ...

        @ferris3.auto_service(endpoint='mobile_only', resource_name='posts', path='posts')
        class MobilePostsService(ferris3.Service):
            ...

    """
    def auto_class_decr(cls):
        if 'resource_name' not in kwargs:
            name = underscore(cls.__name__).replace('_api', '').replace('_service', '')
            kwargs['resource_name'] = name

        if 'path' not in kwargs:
            kwargs['path'] = kwargs['resource_name']

        ep_api = get(endpoint)
        return ep_api.api_class(**kwargs)(cls)

    if cls:
        return auto_class_decr(cls)
    return auto_class_decr


def auto_method(func=None, returns=message_types.VoidMessage, name=None, http_method='POST', **kwargs):
    """
    Uses introspection to automatically configure and expose an API method. This is sugar around `endpoints.method <https://developers.google.com/appengine/docs/python/endpoints/create_api#defining_an_api_method_endpointsmethod>`_.

    The ``returns`` argument is the response message type and is by default ``message_types.VoidMessage`` (an empty response).
    The ``name`` argument is optional and if left out will be set to the name of the function.
    The ``http_method`` argument is ``POST`` by default and can be changed if desired. Note that ``GET`` methods can not accept a request message.
    The remaining ``kwargs`` are passed direction to ``endpoints.method``.

    This decorator uses introspection along with annotation to determine the request message type as well as any query string parameters for the method. This saves you the trouble of having to define a `ResourceContainer <https://developers.google.com/appengine/docs/python/endpoints/create_api#using_resourcecontainer_for_path_or_querystring_arguments>`_.

    Examples::

        # A method that takes nothing and returns nothing.
        @auto_method
        def nothing(self, request):
            pass

        # A method that returns a simple message
        @auto_method(returns=SimpleMessage)
        def simple(self, request):
            return SimpleMessage(content="Hello")

        # A method that takes a simple message as a request
        @auto_method
        def simple_in(self, request=(SimpleMessage,)):
            pass

        # A method that takes 2 required parameters and one optional one.
        @auto_method
        def params(self, request, one=(str,), two=(int,), three=(str, 'I am default')):
            pass

    """
    def auto_api_decr(func):
        func_name = func.__name__ if not name else name
        func = annotated(returns=returns)(func)
        f_annotations = func.__annotations__
        f_args, f_varargs, f_kwargs, f_defaults = inspect.getargspec(func)

        if f_defaults:
            f_args_to_defaults = {f_args[len(f_args) - len(f_defaults) + n]: x for n, x in enumerate(f_defaults)}
        else:
            f_args_to_defaults = {}

        RequestMessage = f_annotations.pop('request', message_types.VoidMessage)
        ResponseMessage = f_annotations.pop('return', message_types.VoidMessage)

        RequestMessageOrContainer, request_args = annotations_to_resource_container(f_annotations, f_args_to_defaults, RequestMessage)

        ep_dec = endpoints.method(
            RequestMessageOrContainer,
            ResponseMessage,
            http_method=http_method,
            name=func_name,
            path=func_name,  #TODO: include required params,
            **kwargs
        )

        def inner(self, request):
            kwargs = {}
            for arg_name in request_args:
                if hasattr(request, arg_name):
                    kwargs[arg_name] = getattr(request, arg_name)
                if arg_name in f_args_to_defaults and kwargs.get(arg_name) is None:
                    kwargs[arg_name] = f_args_to_defaults[arg_name]

            return_val = func(self, request, **kwargs)

            # return voidmessage if the return val is none
            if ResponseMessage == message_types.VoidMessage and return_val is None:
                return message_types.VoidMessage()

            return return_val

        return ep_dec(inner)

    if func:
        return auto_api_decr(func)

    return auto_api_decr


def annotations_to_resource_container(annotations, defaults, RequestMessage):
    args = {}

    if annotations:
        for n, (name, type) in enumerate(annotations.iteritems(), 1):
            required = True if name not in defaults else False
            default = defaults[name] if not required else None

            if type == str:
                args[name] = messages.StringField(n, required=required, default=default)
            if type == int:
                args[name] = messages.IntegerField(n, required=required, default=default)
            if type == bool:
                args[name] = messages.BooleanField(n, required=required, default=default)

    if args:
        return endpoints.ResourceContainer(RequestMessage, **args), args.keys()

    return RequestMessage, args.keys()



from . import settings, template
from google.appengine.api import mail, app_identity
import logging


def send(recipient, subject, body, sender=None, reply_to=None, **kwargs):
    """
    Sends an html email to ``recipient`` with the given ``subject`` and ``body``.

    If sender is none, it's automatically set to ``settings['email_sender']``,
    If the setting is not configured, then the default ``noreply@[appid].appspotmail.com`` is used.

    Any additionally arguments are passed to ``mail.send_mail``, such as headers.
    """
    sender = sender if sender else settings.get('email_sender', None)
    if not sender:
        sender = "noreply@%s.appspotmail.com" % app_identity.get_application_id()
        logging.info("No sender configured, using the default one: %s" % sender)

    res = mail.send_mail(
        sender=sender,
        to=recipient,
        subject=subject,
        body=body,
        html=body,
        reply_to=reply_to if reply_to else sender,
        **kwargs)
    logging.info('Email sent to %s by %s with subject %s and result %s' % (recipient, sender, subject, res))
    return res


def send_template(recipient, subject, template_name, context=None, theme=None, **kwargs):
    """
    Renders a template using :func:`~ferris3.template.render` and sends an email
    in the same way as :func:`send`.

    For example::

        mail.send_template(
            recipient='jondoe@example.com',
            subject='A Test Email',
            template_name='app/email/test.html',
            context={
                'name': 'George'
            }
        )

    Would render the template ``app/email/test.html`` and email the rendered html.
    """
    name = ('email/' + template_name + '.html', template)
    context = context if context else {}
    body = template.render(name, context, theme=theme)
    res = send(recipient, subject, body, **kwargs)
    return res, body



# LICENSE
# =======

# Copyright (c) 2013, Tripp Lilley <tripplilley@gmail.com>
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.

# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.

# - Neither `Agoraplex`, nor the names of its contributors may be used
#   to endorse or promote products derived from this software without
#   specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# anodi
from collections import OrderedDict
import inspect
empty = object()

def returns (annotation):
    """
    Decorator to add ``annotation`` to ``func``'s ``return``
    annotation, as though it were a Python 3 ``-> ...`` annotation.

        >>> from anodi import returns
        >>> @returns(int)
        ... def example ():
        ...    pass
        ...
        >>> example.__annotations__
        {'return': <type 'int'>}
    """
    def annotate (func):
        func.__annotations__ = getattr(func, '__annotations__', {})
        if not annotation is empty:
            func.__annotations__['return'] = annotation
        return func
    return annotate

def annotated (func=None, returns=empty):
    """
    Decorator to treat ``func``'s default args as a combination of
    annotations and default values, migrating the annotations to
    ``func.__annotations__``, leaving only the defaults in
    ``__defaults__``).

    The optional ``returns`` keyword parameter is placed in the
    resulting ``__annotations__`` dict.

    Each default value must be a tuple, ``(annotation, default)``. To
    supply an unannotated parameter with a default value, use the
    ``empty`` marker object. To supply an annotation without a
    default value, use a 1-tuple: ``(annotation,)``.

    Note that the Python 2.x rules prohibiting non-default parameters
    from coming after defaults still apply, but we don't enforce those
    rules. The effect of using the ``(annotation,)`` form *after*
    using the ``(annotation, default)`` form is likely to be
    surprising, at best.

    You may specify an unannotated parameter by using an empty tuple
    as its default value. This is to allow placing unannotated
    parameters after annotated parameters. Ordinarily, this would not
    be allowed, since the annotated parameter would mark the start of
    default values, requiring defaults on all subsequent parameters.

    We do *not* support nested tuple parameters.

    We also don't yet have a way to add annotations to the ``*args``
    or ``**kwargs`` catch-all parameters, since they don't take
    defaults.

    Example:

        >>> from anodi import annotated, empty
        >>> @annotated
        ... def example (a, b, c=(int,), d=(), e=(empty, "hi")):
        ...    pass
        ...
        >>> example.__annotations__
        {'c': <type 'int'>}
        >>> example.__defaults__
        ('hi',)

        >>> @annotated(returns=int)
        ... def example (a, b, c=(int,), d=(), e=(empty, "hi")):
        ...    pass
        ...
        >>> example.__annotations__
        {'c': <type 'int'>, 'return': <type 'int'>}
        >>> example.__defaults__
        ('hi',)

    """

    def annotate (func):
        func.__annotations__ = getattr(func, '__annotations__', OrderedDict())

        if not returns == empty:
            func.__annotations__['return'] = returns

        defaults = func.__defaults__
        if defaults:
            spec = inspect.getargspec(func)
            # ___TODO:___ support *args, **kwargs annotation?

            # extract annotations
            nanno = len(defaults)
            for (i, name) in enumerate(spec.args[-nanno:]):
                if len(defaults[i]) < 1 or defaults[i][0] is empty:
                    continue
                func.__annotations__[name] = defaults[i][0]

            # prune annotations, leaving only defaults
            defaults = tuple((d[1]
                              for d in func.__defaults__
                              if len(d) > 1))
            # use ``None`` if there are no defaults left, since that's
            # how a function without any defaults would come out.
            func.__defaults__ = defaults or None
        return func

    # if we were called without a ``results`` argument, then we're
    # directly decorating ``func``:
    if returns == empty:
        return annotate(func)

    # otherwise, we're indirectly decorating, via ``annotate``:
    return annotate



from __future__ import absolute_import
import httplib2
import logging
import json
import functools
from apiclient import discovery, errors
import hashlib


def build(serviceName, version, credentials):
    """
    Build a Google API client and caches it in the in-process cache. This reduces
    the number of calls to the discovery API as well as making it easy to share
    the client across multiple parts of code with little effort.

    Usage is similar to ``apiclient.discovery.build``, however, instead of passing an http instance
    you just pass in valid credentials and this method will handle constructing an appropriate http instance for you.

    Example::

        credentials = oauth2.build_service_account_credentials(["https://www.googleapis.com/auth/drive"])
        drive = build("drive", "v2", credentials)

    """
    from . import caching
    credentials_hash = hashlib.sha1(credentials.to_json()).hexdigest()
    cache_key = "ferris:google-client-%s-%s-%s" % (serviceName, version, credentials_hash)

    @caching.cache_using_local(cache_key)
    def inner():
        http = httplib2.Http()
        credentials.authorize(http)
        service = discovery.build(serviceName, version, http=http)
        return service

    return inner()


def retry_execute(request):
    """
    Executes the given request from the Google API client and applies the
    appropriate retry policy. This ensures that if your request fails due to
    internal server error or quota denial the request will be automatically
    retried.

    Example::

        request = service.files().list()
        result = retry_execute(request)

    """
    @retries
    def inner():
        return request.execute()
    return inner()


def retries(f):
    """
    Shortcut decorator that uses the appropraite retry policy for dealing with Google APIs.

    Will retry if an HttpError in the 5xx range is raise, but will fail if the error is in the 4xx range.

    This is useful over retry_execute because it can retry an entire function, not just a single request.

    Example::

        @retries
        def rename_file():
            client = build('drive', 'v2')
            client.files().update(fileId="123", data={"name": "Test"}).execute()

    """
    from .retries import retries as ferris_retries

    @functools.wraps(f)
    def inner(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except errors.HttpError as error:
            raise
        except Exception as error:
            logging.error("Non-recoverable exception: %s" % error)
            raise

    r_inner = ferris_retries(max_tries=5, should_retry=apiclient_retry_policy, delay=1, backoff=2)(inner)
    return r_inner


def apiclient_retry_policy(exception):
    if not isinstance(exception, errors.HttpError):
        return False

    try:
        error = json.loads(exception.content)
        error = error.get('error', error)
        code = error.get('code')
        message = error.get('message)')
        reason = error.get('errors', [{}])[0].get('reason')
        if code in (500, 501, 502, 503, 504):
            logging.info("Google returned internal error %s: %s, retrying..." % (code, reason))
            return True
        if code == 403 and reason in ('rateLimitExceeded', 'userRateLimitExceeded'):
            logging.info("Rate limit exceeded, retrying...")
            return True
        elif code == 403 and reason in ('dailyLimitExceeded',):
            logging.error("Uh oh- daily quota limit exceeded! Not retrying")
            return False
        else:
            logging.info("API error %s: %s: %s raised. Not retrying" % (code, reason, message))

    except ValueError:
        logging.error("Failed to parse json from exception: %s" % exception.content)

    return False


def _get_discovery_document(api, api_version, uri_template="https://www.googleapis.com/discovery/v1/apis/{api}/{api_version}/rest", http=None):
    """
    Provides an automatic caching version of the apiclient discovery
    document fetching mechanism using memcache.
    """
    from . import caching
    if not http:
        http = httplib2.Http()

    uri = uri_template.format(api=api, api_version=api_version)

    @caching.cache_using_memcache('gapi-discovery-doc-%s' % uri, 24*60*60)
    def fetch():
        r, c = http.request(uri)
        return r, c

    r, c = fetch()

    return c


def _patch_discovery():
    original_build = discovery.build

    def patched_build(serviceName, version, http=None, **kwargs):
        doc = _get_discovery_document(serviceName, version, http=http)
        return discovery.build_from_document(doc, http=http, **kwargs)

    discovery.build = patched_build
    setattr(discovery, '_build', original_build)


_patch_discovery()



from setuptools import setup

VERSION = '3.0.0-alpha4'

setup(
    name="Ferris",
    version=VERSION,
    author='Jon Parrott / Cloud Sherpas',
    author_email='jjramone13@gmail.com',
    maintainer='Jon Parrott / Cloud Sherpas',
    maintainer_email='jonathan.parrott@cloudsherpas.com',
    description='A framework for creating APIs on Google App Engine',
    url='https://bitbucket.org/cloudsherpas/ferris3',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    packages=['ferris3'],
    install_requires=[
        'google-api-python-client',
        'protopigeon'
    ],
)



from __future__ import absolute_import

from . import ndb, messages, template, settings, hvild
from protorpc.remote import Service
from protorpc.message_types import VoidMessage
from .endpoints import auto_method, auto_service
from .tool_chain import ToolChain
from .endpoints import default as default_endpoint
from .ndb import Model, Behavior
from protopigeon import model_message, list_message
from endpoints import get_current_user, NotFoundException, BadRequestException
